Require Import mtac.
Require Import Lists.List.

Record rel := mkrel {
  left : Type;
  right : Type;
  relation : left -> right -> Type
}.

Definition env := list (Type * rel).

Definition Notfound : Exception. exact exception. Qed.

Definition lookup : Type -> env -> M rel :=
  fun T =>
    mfix loop (rho : env) : M rel :=
    mmatch rho return M rel with
    | nil => raise Notfound
    | [(R:rel) rest] (T, R) :: rest => ret R
    | [x (rest : env)] x :: rest => loop rest
   end.


Program
Definition translate : Type -> env -> M rel :=
  mfix translate (T : Type) (rho : env) :=
    mmatch T with
    | [A] (forall x:Type, A x) =>
      nu (x x1 x2:Type) (xr : x1 -> x2 -> Type),
      Aresult <- translate (A x) (cons (x, (mkrel x1 x2 xr)) rho);
      let (A1, A2, AR) := (Aresult : rel) in
      A1' <- abs_eq x1 A1;
      A2' <- abs_eq x2 A2;
      AR0 <- @abs _ (fun _=>_) (xr : x1 -> x2 -> Type) (AR : A1 -> A2 -> Type);
      _
   | [A B] (A -> B) =>
     Aresult <- translate A rho;
     Bresult <- translate B rho;
     let (A1, A2, AR) := Aresult in
     let (B1, B2, BR) := Bresult in
     ret (mkrel (A1 -> B1)
                (A2 -> B2)
                (fun f1 f2 =>
                   forall (a1 : A1) (a2 : A2),
                     (AR a1 a2) -> (BR (f1 a1) (f2 a2))))
   | [x] x => lookup x rho 
    end. 
Next Obligation.
intros.
destruct A1' as [A1' pf1].
destruct A2' as [A2' pf2].
rewrite <- pf1, <- pf2 in AR0.
refine (
  AR1 <- abs x2 AR0;
  AR2 <- @abs _ (fun x1'=>forall x2', (x1' -> x2' -> Type) -> A1' x1' -> A2' x2' -> Type) x1 AR1;
  ret (mkrel (forall x1, A1' x1)
                 (forall x2, A2' x2)
                 (fun a1 a2 => forall (x1 x2 : Type) (xr : x1 -> x2 -> Type),
                                 AR2 x1 x2 xr (a1 x1) (a2 x2)))).                
Defined.

Definition rel_t_arr_t := 
  run (translate (Type -> Type) ((Type, mkrel Type Type (fun l r=>l = r)) :: nil)).
Eval compute in rel_t_arr_t.

Definition rel_x_arr_x := 
  run (translate (forall x:Type, x -> x) ((Type, mkrel Type Type (fun l r=>l = r)) :: nil)).
Eval compute in rel_x_arr_x.

